"""
CATS maintenance commands hardware object.

Functionality in addition to sample-transfer functionality: power control,
lid control, error-recovery commands, ...

Derived from Michael Hellmig's implementation for the BESSY CATS sample changer
 -add more controls, including Regulation Off, Gripper Dry/Open/Close, Reset Memory, Set Sample On Diff
 -add CATS dewar layout

Vicente Rey - add support for ISARA Model

"""

import logging
import time

import gevent
from PyTango import DeviceProxy

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import task

__author__ = "Jie Nan"
__credits__ = ["The MxCuBE collaboration"]

__email__ = "jie.nan@maxlab.lu.se"
__status__ = "Alpha"

TOOL_FLANGE, TOOL_UNIPUCK, TOOL_SPINE, TOOL_PLATE, TOOL_LASER, TOOL_DOUBLE_GRIPPER = (
    0,
    1,
    2,
    3,
    4,
    5,
)

TOOL_TO_STR = {
    "Flange": TOOL_FLANGE,
    "Unipuck": TOOL_UNIPUCK,
    "EMBL": TOOL_SPINE,
    "Rotat": TOOL_SPINE,
    "Plate": TOOL_PLATE,
    "Laser": TOOL_LASER,
    "Double": TOOL_DOUBLE_GRIPPER,
}


class CatsMaint(HardwareObject):
    __TYPE__ = "CATS"
    NO_OF_LIDS = 3

    """
    Actual implementation of the CATS Sample Changer, MAINTENANCE COMMANDS ONLY
    BESSY BL14.1 installation with 3 lids
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._state = None
        self._running = None
        self._powered = None
        self._toolopen = None
        self._message = None
        self._regulating = None
        self._lid1state = None
        self._lid2state = None
        self._lid3state = None
        self._charging = None

    def init(self):
        self.cats_device = DeviceProxy(self.tangoname)

        try:
            self.cats_model = self.cats_device.read_attribute("CatsModel").value
        except Exception:
            self.cats_model = "CATS"

        if self.is_isara():
            self.nb_of_lids = 1
        else:
            self.nb_of_lids = 3

        self._chnState = self.add_channel(
            {
                "type": "tango",
                "name": "_chnState",
                "tangoname": self.tangoname,
                "polling": 1000,
            },
            "State",
        )

        self._chnPathRunning = self.add_channel(
            {
                "type": "tango",
                "name": "_chnPathRunning",
                "tangoname": self.tangoname,
                "polling": 1000,
            },
            "PathRunning",
        )
        self._chnPowered = self.add_channel(
            {
                "type": "tango",
                "name": "_chnPowered",
                "tangoname": self.tangoname,
                "polling": 1000,
            },
            "Powered",
        )
        self._chnMessage = self.add_channel(
            {
                "type": "tango",
                "name": "_chnMessage",
                "tangoname": self.tangoname,
                "polling": 1000,
            },
            "Message",
        )
        self._chnToolOpenClose = self.add_channel(
            {
                "type": "tango",
                "name": "_chnToolOpenClose",
                "tangoname": self.tangoname,
                "polling": 1000,
            },
            "di_ToolOpen",
        )
        self._chnLN2Regulation = self.add_channel(
            {
                "type": "tango",
                "name": "_chnLN2Regulation",
                "tangoname": self.tangoname,
                "polling": 1000,
            },
            "LN2Regulating",
        )
        self._chnBarcode = self.add_channel(
            {
                "type": "tango",
                "name": "_chnBarcode",
                "tangoname": self.tangoname,
                "polling": 1000,
            },
            "Barcode",
        )

        self._chnLid1State = self.add_channel(
            {
                "type": "tango",
                "name": "_chnLid1State",
                "tangoname": self.tangoname,
                "polling": 1000,
            },
            "di_Lid1Open",
        )
        self._chnLid1State.connect_signal("update", self._update_lid1_state)

        if self.nb_of_lids > 1:
            self._chnLid2State = self.add_channel(
                {
                    "type": "tango",
                    "name": "_chnLid2State",
                    "tangoname": self.tangoname,
                    "polling": 1000,
                },
                "di_Lid2Open",
            )
            self._chnLid2State.connect_signal("update", self._update_lid2_state)

        if self.nb_of_lids > 2:
            self._chnLid3State = self.add_channel(
                {
                    "type": "tango",
                    "name": "_chnLid3State",
                    "tangoname": self.tangoname,
                    "polling": 1000,
                },
                "di_Lid3Open",
            )
            self._chnLid3State.connect_signal("update", self._update_lid3_state)

        self._chnState.connect_signal("update", self._update_state)
        self._chnPathRunning.connect_signal("update", self._update_running_state)
        self._chnPowered.connect_signal("update", self._update_powered_state)
        self._chnToolOpenClose.connect_signal("update", self._update_tool_state)
        self._chnMessage.connect_signal("update", self._update_message)
        self._chnLN2Regulation.connect_signal("update", self._update_regulation_state)
        self._chnBarcode.connect_signal("update", self._updateBarcode)

        self._chnCurrentTool = self.add_channel(
            {"type": "tango", "name": "_chnCurrentTool", "tangoname": self.tangoname},
            "Tool",
        )
        #
        self._cmdPowerOn = self.add_command(
            {"type": "tango", "name": "_cmdPowerOn", "tangoname": self.tangoname},
            "powerOn",
        )
        self._cmdPowerOff = self.add_command(
            {"type": "tango", "name": "_cmdPowerOff", "tangoname": self.tangoname},
            "powerOff",
        )
        self._cmdOpenTool = self.add_command(
            {"type": "tango", "name": "_cmdOpenTool", "tangoname": self.tangoname},
            "opentool",
        )
        self._cmdCloseTool = self.add_command(
            {"type": "tango", "name": "_cmdCloseTool", "tangoname": self.tangoname},
            "closetool",
        )
        self._cmdMagnetOn = self.add_command(
            {"type": "tango", "name": "_cmdMagnetOn", "tangoname": self.tangoname},
            "magnetOn",
        )
        self._cmdMagnetOff = self.add_command(
            {"type": "tango", "name": "_cmdMagnetOff", "tangoname": self.tangoname},
            "magnetOff",
        )

        # LIDs
        self._cmdOpenLid1 = self.add_command(
            {"type": "tango", "name": "_cmdOpenLid1", "tangoname": self.tangoname},
            "openlid1",
        )
        self._cmdCloseLid1 = self.add_command(
            {"type": "tango", "name": "_cmdCloseLid1", "tangoname": self.tangoname},
            "closelid1",
        )

        if self.nb_of_lids > 1:
            self._cmdOpenLid2 = self.add_command(
                {"type": "tango", "name": "_cmdOpenLid1", "tangoname": self.tangoname},
                "openlid2",
            )
            self._cmdCloseLid2 = self.add_command(
                {"type": "tango", "name": "_cmdCloseLid1", "tangoname": self.tangoname},
                "closelid2",
            )

        if self.nb_of_lids > 2:
            self._cmdOpenLid3 = self.add_command(
                {"type": "tango", "name": "_cmdOpenLid1", "tangoname": self.tangoname},
                "openlid3",
            )
            self._cmdCloseLid3 = self.add_command(
                {"type": "tango", "name": "_cmdCloseLid1", "tangoname": self.tangoname},
                "closelid3",
            )

        self._cmdRegulOn = self.add_command(
            {"type": "tango", "name": "_cmdRegulOn", "tangoname": self.tangoname},
            "regulon",
        )
        self._cmdRegulOff = self.add_command(
            {"type": "tango", "name": "_cmdRegulOff", "tangoname": self.tangoname},
            "reguloff",
        )

        self._cmdToolOpen = self.add_command(
            {"type": "tango", "name": "_cmdToolOpen", "tangoname": self.tangoname},
            "opentool",
        )
        self._cmdToolClose = self.add_command(
            {"type": "tango", "name": "_cmdToolClose", "tangoname": self.tangoname},
            "closetool",
        )

        # Paths
        self._cmdAbort = self.add_command(
            {"type": "tango", "name": "_cmdAbort", "tangoname": self.tangoname}, "abort"
        )
        self._cmdDry = self.add_command(
            {"type": "tango", "name": "_cmdDry", "tangoname": self.tangoname}, "dry"
        )
        self._cmdSafe = self.add_command(
            {"type": "tango", "name": "_cmdSafe", "tangoname": self.tangoname}, "safe"
        )
        self._cmdHome = self.add_command(
            {"type": "tango", "name": "_cmdHome", "tangoname": self.tangoname}, "home"
        )
        self._cmdSoak = self.add_command(
            {"type": "tango", "name": "_cmdSoak", "tangoname": self.tangoname}, "soak"
        )
        self._cmdBack = self.add_command(
            {"type": "tango", "name": "_cmdBack", "tangoname": self.tangoname}, "back"
        )
        self._cmdCalibration = self.add_command(
            {"type": "tango", "name": "_cmdCalibration", "tangoname": self.tangoname},
            "toolcalibration",
        )

        self._cmdClearMemory = self.add_command(
            {"type": "tango", "name": "_cmdClearMemory", "tangoname": self.tangoname},
            "clear_memory",
        )
        self._cmdReset = self.add_command(
            {"type": "tango", "name": "_cmdReset", "tangoname": self.tangoname}, "reset"
        )
        self._cmdResetParameters = self.add_command(
            {
                "type": "tango",
                "name": "_cmdResetParameters",
                "tangoname": self.tangoname,
            },
            "reset_parameters",
        )

        self._cmdRecoverFailure = self.add_command(
            {
                "type": "tango",
                "name": "_cmdRecoverFailure",
                "tangoname": self.tangoname,
            },
            "recoverFailure",
        )

        self._cmdResetMotion = self.add_command(
            {"type": "tango", "name": "_cmdResetMotion", "tangoname": self.tangoname},
            "resetmotion",
        )

        self._cmdSetOnDiff = self.add_command(
            {"type": "tango", "name": "_cmdSetOnDiff", "tangoname": self.tangoname},
            "setondiff",
        )
        self._cmdSetOnTool = self.add_command(
            {"type": "tango", "name": "_cmdSetOnTool", "tangoname": self.tangoname},
            "settool",
        )
        self._cmdSetOnTool2 = self.add_command(
            {"type": "tango", "name": "_cmdSetOnTool2", "tangoname": self.tangoname},
            "settool2",
        )

        self.state_actions = {
            "power": {
                "in_open": self._cmdPowerOn,
                "out_close": self._cmdPowerOff,
                "state": self._chnPowered,
            }
        }

    def is_isara(self):
        return self.cats_model == "ISARA"

    def is_cats(self):
        return self.cats_model != "ISARA"

    def get_current_tool(self):
        current_value = self._chnCurrentTool.get_value()

        tool = TOOL_TO_STR.get(current_value, None)

        return tool

    ################################################################################

    def back_traj(self):
        """
        Moves a sample from the gripper back into the dewar to its logged position.
        """
        return self._execute_task(False, self._do_back)

    def safe_traj(self):
        """
        Safely Moves the robot arm and the gripper to the home position
        """
        return self._execute_task(False, self._do_safe)

    def _do_abort(self):
        """
        Launch the "abort" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        self._cmdAbort()

    def _do_home(self):
        """
        Launch the "abort" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        tool = self.get_current_tool()
        self._cmdHome(tool)

    def _do_reset(self):
        """
        Launch the "reset" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        logging.getLogger("HWR").debug("CatsMaint. doing reset")
        return
        self._cmdReset()

    def _do_reset_memory(self):
        """
        Launch the "reset memory" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        self._cmdClearMemory()
        time.sleep(1)
        self._cmdResetParameters()
        time.sleep(1)

    def _do_reset_motion(self):
        """
        Launch the "reset_motion" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        self._cmdResetMotion()

    def _do_recover_failure(self):
        """
        Launch the "recoverFailure" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        self._cmdRecoverFailure()

    def _do_calibration(self):
        """
        Launch the "toolcalibration" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        tool = self.get_current_tool()
        self._cmdCalibration(tool)

    def _do_open_tool(self):
        """
        Launch the "opentool" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        self._cmdOpenTool()

    def _do_close_tool(self):
        """
        Launch the "closetool" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        self._cmdCloseTool()

    def _do_dry_gripper(self):
        """
        Launch the "dry" command on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        tool = self.get_current_tool()
        self._cmdDry(tool)

    def _do_set_on_diff(self, sample):
        """
        Launch the "setondiff" command on the CATS Tango DS, an example of sample value is 2:05

        :returns: None
        :rtype: None
        """

        if sample is None:
            raise Exception("No sample selected")
        else:
            str_tmp = str(sample)
            sample_tmp = str_tmp.split(":")
            # calculate CATS specific lid/sample number
            lid = (int(sample_tmp[0]) - 1) / 3 + 1
            puc_pos = ((int(sample_tmp[0]) - 1) % 3) * 10 + int(sample_tmp[1])
            argin = [str(lid), str(puc_pos), "0"]
            logging.getLogger().info("to SetOnDiff %s", argin)
            self._execute_server_task(self._cmdSetOnDiff, argin)

    def _do_back(self):
        """
        Launch the "back" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        tool = self.get_current_tool()
        argin = [str(tool), "0"]  # to send string array with two arg...
        self._execute_server_task(self._cmdBack, argin)

    def _do_safe(self):
        """
        Launch the "safe" trajectory on the CATS Tango DS

        :returns: None
        :rtype: None
        """
        argin = self.get_current_tool()
        self._execute_server_task(self._cmdSafe, argin)

    def _do_power_state(self, state=False):
        """
        Switch on CATS power if >state< == True, power off otherwise

        :returns: None
        :rtype: None
        """
        logging.getLogger("HWR").debug("   running power state command ")
        if state:
            self._cmdPowerOn()
        else:
            self._cmdPowerOff()

        self.do_state_action("power", state)

    def _do_enable_regulation(self):
        """
        Switch on CATS regulation

        :returns: None
        :rtype: None
        """
        self._cmdRegulOn()

    def _do_disable_regulation(self):
        """
        Switch off CATS regulation

        :returns: None
        :rtype: None
        """
        self._cmdRegulOff()

    def _do_lid1_state(self, state=True):
        """
        Opens lid 1 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        if state:
            self._execute_server_task(self._cmdOpenLid1)
        else:
            self._execute_server_task(self._cmdCloseLid1)

    def _do_lid2_state(self, state=True):
        """
        Opens lid 2 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        if state:
            self._execute_server_task(self._cmdOpenLid2)
        else:
            self._execute_server_task(self._cmdCloseLid2)

    def _do_lid3_state(self, state=True):
        """
        Opens lid 3 if >state< == True, closes the lid otherwise

        :returns: None
        :rtype: None
        """
        if state:
            self._execute_server_task(self._cmdOpenLid3)
        else:
            self._execute_server_task(self._cmdCloseLid3)

    def _do_magnet_on(self):
        self._execute_server_task(self._cmdMagnetOn)

    def _do_magnet_off(self):
        self._execute_server_task(self._cmdMagnetOff)

    def _do_tool_open(self):
        self._execute_server_task(self._cmdToolOpen)

    def _do_tool_close(self):
        self._execute_server_task(self._cmdToolClose)

    # ########################          PROTECTED          #########################

    def _execute_task(self, wait, method, *args):
        ret = self._run(method, wait=False, *args)
        if wait:
            return ret.get()
        else:
            return ret

    @task
    def _run(self, method, *args):
        exception = None
        ret = None
        try:
            ret = method(*args)
        except Exception as ex:
            exception = ex
        if exception is not None:
            raise exception
        return ret

    # ########################           PRIVATE           #########################

    def _update_running_state(self, value):
        self._running = value
        self.emit("runningStateChanged", (value,))
        self._update_global_state()

    def _update_powered_state(self, value):
        self._powered = value
        self.emit("powerStateChanged", (value,))
        self._update_global_state()

    def _update_tool_state(self, value):
        self._toolopen = value
        self.emit("toolStateChanged", (value,))
        self._update_global_state()

    def _update_message(self, value):
        self._message = value
        self.emit("messageChanged", (value,))
        self._update_global_state()

    def _update_regulation_state(self, value):
        self._regulating = value
        self.emit("regulationStateChanged", (value,))
        self._update_global_state()

    def _updateBarcode(self, value):
        self._barcode = value
        self.emit("barcodeChanged", (value,))

    def _update_state(self, value):
        self._state = value
        self._update_global_state()

    def _update_lid1_state(self, value):
        self._lid1state = value
        self.emit("lid1StateChanged", (value,))
        self._update_global_state()

    def _update_lid2_state(self, value):
        self._lid2state = value
        self.emit("lid2StateChanged", (value,))
        self._update_global_state()

    def _update_lid3_state(self, value):
        self._lid3state = value
        self.emit("lid3StateChanged", (value,))
        self._update_global_state()

    def _update_operation_mode(self, value):
        self._charging = not value

    def _update_global_state(self):
        state_dict, cmd_state, message = self.get_global_state()
        self.emit("globalStateChanged", (state_dict, cmd_state, message))

    def get_global_state(self):
        """
        Update clients with a global state that
        contains different:

        - first param (state_dict):
            collection of state bits

        - second param (cmd_state):
            list of command identifiers and the
            status of each of them True/False
            representing whether the command is
            currently available or not

        - message
            a message describing current state information
            as a string
        """
        _ready = str(self._state) in ("READY", "ON")

        if self._running:
            state_str = "MOVING"
        elif not (self._powered) and _ready:
            state_str = "DISABLED"
        elif _ready:
            state_str = "READY"
        else:
            state_str = str(self._state)

        state_dict = {
            "toolopen": self._toolopen,
            "powered": self._powered,
            "running": self._running,
            "regulating": self._regulating,
            "lid1": self._lid1state,
            "lid2": self._lid2state,
            "lid3": self._lid3state,
            "state": state_str,
        }

        cmd_state = {
            "powerOn": (not self._powered) and _ready,
            "powerOff": (self._powered) and _ready,
            "regulon": (not self._regulating) and _ready,
            "openlid1": (not self._lid1state) and self._powered and _ready,
            "closelid1": self._lid1state and self._powered and _ready,
            "dry": (not self._running) and self._powered and _ready,
            "soak": (not self._running) and self._powered and _ready,
            "home": (not self._running) and self._powered and _ready,
            "back": (not self._running) and self._powered and _ready,
            "safe": (not self._running) and self._powered and _ready,
            "clear_memory": True,
            "reset": True,
            "abort": True,
        }

        message = self._message

        return state_dict, cmd_state, message

    def get_cmd_info(self):
        """return information about existing commands for this object
        the information is organized as a list
        with each element contains
        [ cmd_name,  display_name, category ]
        """
        """ [cmd_id, cmd_display_name, nb_args, cmd_category, description ] """
        cmd_list = [
            [
                "Power",
                [
                    ["powerOn", "PowerOn", "Switch Power On"],
                    ["powerOff", "PowerOff", "Switch Power Off"],
                    ["regulon", "Regulation On", "Swich LN2 Regulation On"],
                ],
            ],
            [
                "Lid",
                [
                    ["openlid1", "Open Lid", "Open Lid"],
                    ["closelid1", "Close Lid", "Close Lid"],
                ],
            ],
            [
                "Actions",
                [
                    ["home", "Home", "Actions", "Home (trajectory)"],
                    ["dry", "Dry", "Actions", "Dry (trajectory)"],
                    ["soak", "Soak", "Actions", "Soak (trajectory)"],
                ],
            ],
            [
                "Recovery",
                [
                    [
                        "clear_memory",
                        "Clear Memory",
                        "Clear Info in Robot Memory "
                        " (includes info about sample on Diffr)",
                    ],
                    ["reset", "Reset Message", "Reset Cats State"],
                    ["back", "Back", "Reset Cats State"],
                    ["safe", "Safe", "Reset Cats State"],
                ],
            ],
            ["Abort", [["abort", "Abort", "Abort Execution of Command"]]],
        ]
        return cmd_list

    def _execute_server_task(self, method, *args):
        task_id = method(*args)
        ret = None
        # introduced wait because it takes some time before the attribute PathRunning is set
        # after launching a transfer
        # after setting refresh in the Tango DS to 0.1 s a wait of 1s is enough
        time.sleep(1.0)
        while str(self._chnPathRunning.get_value()).lower() == "true":
            gevent.sleep(0.1)
        ret = True
        return ret

    def send_command(self, cmd_name, args=None):
        #
        lid = 1
        toolcal = 0
        tool = self.get_current_tool()

        if cmd_name in ["dry", "safe", "home"]:
            if tool is not None:
                args = [tool]
            else:
                raise Exception("Cannot detect type of TOOL in Cats. Command ignored")

        if cmd_name == "soak":
            if tool in [TOOL_DOUBLE, TOOL_UNIPUCK]:
                args = [str(tool), str(lid)]
            else:
                raise Exception("Can SOAK only when UNIPUCK tool is mounted")

        if cmd_name == "back":
            if tool is not None:
                args = [tool, toolcal]
            else:
                raise Exception("Cannot detect type of TOOL in Cats. Command ignored")

        cmd = getattr(self.cats_device, cmd_name)

        try:
            if args is not None:
                if len(args) > 1:
                    ret = cmd(map(str, args))
                else:
                    ret = cmd(*args)
            else:
                ret = cmd()
            return ret
        except Exception as exc:
            import traceback

            traceback.print_exc()
            msg = exc[0].desc
            raise Exception(msg)


def test_hwo(hwo):
    print((hwo.get_current_tool()))



"""
MAXIV Session hardware object.

Adapting from original Session.py to adapt the names of data directories
"""

import logging
import os
import socket
import time

try:
    from sdm import (
        Visitor,
        storage,
    )
except Exception:
    raise Exception("Cannot import SDM library.")

from mxcubecore.HardwareObjects.Session import Session
from mxcubecore.model import queue_model_objects


class MaxIVSession(Session):
    def __init__(self, name):
        Session.__init__(self, name)

    # Framework-2 method, inherited from HardwareObject and called
    # by the framework after the object has been initialized.
    def init(self):
        self.default_precision = 4
        self.login = ""
        self.is_commissioning = False
        self.synchrotron_name = self.get_property("synchrotron_name")
        self.endstation_name = self.get_property("endstation_name").lower()
        self.suffix = self["file_info"].get_property("file_suffix")
        self.base_directory = self["file_info"].get_property("base_directory")
        self.base_process_directory = self["file_info"].get_property(
            "processed_data_base_directory"
        )

        self.raw_data_folder_name = self["file_info"].get_property(
            "raw_data_folder_name"
        )

        self.processed_data_folder_name = self["file_info"].get_property(
            "processed_data_folder_name"
        )

        try:
            self.in_house_users = self.get_property("inhouse_users").split(",")
        except Exception:
            self.in_house_users = []

        try:
            domain = socket.getfqdn().split(".")
            self.email_extension = ".".join((domain[-2], domain[-1]))
        except (TypeError, IndexError):
            pass

        archive_base_directory = self["file_info"].get_property(
            "archive_base_directory"
        )
        if archive_base_directory:
            queue_model_objects.PathTemplate.set_archive_path(
                archive_base_directory, self["file_info"].get_property("archive_folder")
            )

        queue_model_objects.PathTemplate.set_path_template_style(self.synchrotron_name)
        queue_model_objects.PathTemplate.set_data_base_path(self.base_directory)

        self.commissioning_fake_proposal = {
            "Laboratory": {
                "address": None,
                "city": "Lund",
                "laboratoryId": 312171,
                "name": "Lund University",
            },
            "Person": {
                "familyName": "commissioning",
                "givenName": "",
                "laboratoryId": 312171,
                "login": "",
                "personId": 0,
            },
            "Proposal": {
                "code": "MX",
                "number": time.strftime("%Y"),
                "proposalId": "0",
                "timeStamp": time.strftime("%Y%m%d"),
                "title": "Commissioning Proposal",
                "type": "MX",
            },
            "Session": [
                {
                    "is_inhouse": True,
                    "beamlineName": "BioMAX",
                    "comments": "Fake session for commissioning",
                    "endDate": "2027-12-31 23:59:59",
                    "nbShifts": 100,
                    "proposalId": "0",
                    "scheduled": 0,
                    "sessionId": 0,
                    "startDate": "2016-01-01 00:00:00",
                }
            ],
        }

    def set_in_commissioning(self, proposal_info):
        self.proposal_code = proposal_info["Proposal"]["code"]
        self.proposal_number = proposal_info["Proposal"]["number"]
        self.is_commissioning = True

    def get_proposal(self):
        """
        :returns: The proposal, 'local-user' if no proposal is
                  available
        :rtype: str
        """
        proposal = "local-user"

        if self.proposal_code and self.proposal_number:
            if self.proposal_code == "ifx":
                self.proposal_code = "fx"

            proposal = str(self.proposal_number)

        return proposal

    def get_base_data_directory(self):
        """
        Returns the base data directory taking the 'contextual'
        information into account, such as if the current user
        is inhouse.

        :returns: The base data path.
        :rtype: str
        """
        if self.session_start_date:
            start_time = self.session_start_date.split(" ")[0].replace("-", "")
        else:
            start_time = time.strftime("%Y%m%d")

        if not self.is_commissioning:
            # /data/visitors/biomax/prop/visit/
            # /data/(user-type)/(beamline)/(proposal)/(visit)/raw
            directory = os.path.join(
                self.base_directory,
                "visitors",  # self.get_user_category(self.login),
                self.beamline_name.lower(),
                self.get_proposal(),
                start_time,
            )
        else:
            # /data/staff/biomax/commissioning/date
            directory = os.path.join(
                self.base_directory,
                "staff",
                self.beamline_name.lower(),
                "commissioning",
                time.strftime("%Y%m%d"),
            )

        logging.getLogger("HWR").info(
            "[MAX IV Session] Data directory for proposal %s: %s"
            % (self.get_proposal(), directory)
        )
        return directory

    def prepare_directories(self, proposal_info):
        self.login = proposal_info["Person"]["login"]
        start_time = proposal_info.get("Session")[0].get("startDate")

        logging.getLogger("HWR").info(
            "[MAX IV Session] Preparing Data directory for proposal %s" % proposal_info
        )
        if start_time:
            start_date = start_time.split(" ")[0].replace("-", "")
        else:
            start_date = time.strftime("%Y%m%d")

        self.set_session_start_date(start_date)

        try:
            self.storage = storage.Storage(
                self.get_user_category(self.login), self.endstation_name
            )
        except Exception as ex:
            print(ex)
            # this creates the path for the data and ensures proper permissions.
            # e.g. /data/visitors/biomax/<proposal>/<visit>/{raw, process}
        if self.is_commissioning:
            group = self.beamline_name.lower()
        else:
            group = self.storage.get_proposal_group(self.proposal_number)
        try:
            self.storage.create_path(
                self.proposal_number, group, self.get_session_start_date()
            )

            proposal_path = "{0}/{1}".format(
                self.storage.beamline_path, self.proposal_number
            )
            logging.getLogger("HWR").info(
                "[MAX IV Session] SDM Data directory created: %s" % proposal_path
            )
        except Exception as ex:
            msg = "[MAX IV Session] SDM Data directory creation failed. %s" % ex
            logging.getLogger("HWR").error(msg)
            raise Exception(msg)

    def is_inhouse(self, user, code=None):
        """
        Determines if a given user is considered to be inhouse.
        :param login: username
        :type login: str
        :returns: True if the user is inhouse, otherwise False.
        :rtype: bool
        """
        if user in self.in_house_users:
            return True
        else:
            return False

    def get_user_category(self, user):
        # 'staff','visitors', 'proprietary'
        # missing industrial users
        if self.is_inhouse(user):
            user_category = "staff"
        else:
            user_category = "visitors"
        return user_category



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import tempfile
from csv import reader
from datetime import datetime
from random import random
from time import sleep

import gevent
import numpy
import numpy as np
import tine
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure
from scipy.interpolate import interp1d

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects import SimpleHTML

# try:
#    import pdfkit
# except Exception:
#    logging.getLogger("HWR").warning("pdfkit not available")



__credits__ = ["EMBL Hamburg"]
__version__ = "2.3."
__category__ = "General"


TEST_DICT = {
    "ppu": "PPU control",
    "focusing": "Focusing modes",
    "centerbeam": "Beam centering",
}

TEST_COLORS_TABLE = {False: "#FFCCCC", True: "#CCFFCC"}
TEST_COLORS_FONT = {False: "#FE0000", True: "#007800"}


class EMBLBeamlineTest(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.ready_event = None
        self.user_clicked_event = None
        self.test_queue_dict = None
        self.current_test_procedure = None
        self.beamline_name = None
        self.test_directory = None
        self.test_filename = None

        self.available_tests_dict = {}
        self.startup_test_list = []
        self.results_list = None
        self.results_html_list = None

        self.bl_hwobj = None
        self.beam_focusing_hwobj = None

    def init(self):
        """Reads config xml, initiates all necessary hwobj, channels and cmds
        """
        self.ready_event = gevent.event.Event()

        self.bl_hwobj = self.get_object_by_role("beamline_setup")
        self.test_filename = "mxcube_test_report"

        try:
            for test in eval(self.get_property("available_tests", "[]")):
                self.available_tests_dict[test] = TEST_DICT[test]
        except Exception:
            logging.getLogger("HWR").debug(
                "BeamlineTest: No test define in xml. "
                + "Setting all tests as available."
            )
        if self.available_tests_dict is None:
            self.available_tests_dict = TEST_DICT

        if self.get_property("startup_tests"):
            self.startup_test_list = eval(self.get_property("startup_tests"))

        if self.get_property("run_tests_at_startup") == True:
            gevent.spawn_later(5, self.start_test_queue, self.startup_test_list)

    def start_test_queue(self, test_list, create_report=True):
        """Runs a list of tests

        :param test_list: list of tests
        :type test_list: list of str
        :param create_report: create html and pdf reports
        :type create_report: bool
        """
        if create_report:
            try:
                logging.getLogger("HWR").debug(
                    "BeamlineTest: Creating directory %s" % self.test_directory
                )
                if not os.path.exists(self.test_directory):
                    os.makedirs(self.test_directory)

                logging.getLogger("HWR").debug(
                    "BeamlineTest: Creating source directory %s" % self.test_directory
                )
                if not os.path.exists(self.test_directory):
                    os.makedirs(self.test_directory)
            except Exception:
                logging.getLogger("HWR").warning(
                    "BeamlineTest: Unable to create test directories"
                )
                return

        self.results_list = []
        self.results_html_list = []
        for test_index, test_name in enumerate(test_list):
            test_method_name = "test_" + test_name.lower()
            if hasattr(self, test_method_name):
                if test_name in TEST_DICT:
                    logging.getLogger("HWR").debug(
                        "BeamlineTest: Executing test %s (%s)"
                        % (test_name, TEST_DICT[test_name])
                    )

                    progress_info = {
                        "progress_total": len(test_list),
                        "progress_msg": "executing %s" % TEST_DICT[test_name],
                    }
                    self.emit("testProgress", (test_index, progress_info))

                    start_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                    # self.current_test_procedure = gevent.spawn(\
                    test_result = getattr(self, test_method_name)()
                    # test_result = self.current_test_procedure.get()

                    self.ready_event.wait()
                    self.ready_event.clear()
                    end_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                    self.results_list.append(
                        {
                            "short_name": test_name,
                            "full_name": TEST_DICT[test_name],
                            "result_bit": test_result.get("result_bit", False),
                            "result_short": test_result.get("result_short", ""),
                            "start_time": start_time,
                            "end_time": end_time,
                        }
                    )

                    self.results_html_list.append(
                        "<h2 id=%s>%s</h2>" % (test_name, TEST_DICT[test_name])
                    )
                    self.results_html_list.append("Started: %s<br>" % start_time)
                    self.results_html_list.append("Ended: %s<br>" % end_time)
                    if test_result.get("result_short"):
                        self.results_html_list.append(
                            "<h3><font color=%s>Result : %s</font></h3>"
                            % (
                                TEST_COLORS_FONT[test_result["result_bit"]],
                                test_result["result_short"],
                            )
                        )
                    if len(test_result.get("result_details", [])) > 0:
                        self.results_html_list.append("<h3>Detailed results:</h3>")
                        self.results_html_list.extend(
                            test_result.get("result_details", [])
                        )
                    self.emit("progressStop", ())
            else:
                msg = (
                    "<h2><font color=%s>Execution method %s "
                    + "for the test %s does not exist</font></h3>"
                )
                self.results_html_list.append(
                    msg
                    % (TEST_COLORS_FONT[False], test_method_name, TEST_DICT[test_name])
                )
                logging.getLogger("HWR").error(
                    "BeamlineTest: Test method " + "%s not available" % test_method_name
                )
            self.results_html_list.append("</p>\n<hr>")

        html_filename = None
        if create_report:
            html_filename = (
                os.path.join(self.test_directory, self.test_filename) + ".html"
            )
            self.generate_report()

        self.emit("testFinished", html_filename)

    def focusing_mode_changed(self, focusing_mode, beam_size):
        """Reemits focusing changed signal

        :param focusing_mode: focusing mode
        :type focusing_mode: str
        :param beam_size: beam size in microns
        :type beam_size: list with two int
        """
        self.emit("focusingModeChanged", focusing_mode, beam_size)

    def get_focus_mode_names(self):
        """Returns available focusing mode names"""
        if self.beam_focusing_hwobj:
            return self.beam_focusing_hwobj.get_focus_mode_names()

    def get_focus_motors(self):
        """Returns focusing motor hwobj"""
        if self.beam_focusing_hwobj is not None:
            return self.beam_focusing_hwobj.get_focus_motors()

    def get_focus_mode(self):
        """Returns active focusing mode"""
        if self.beam_focusing_hwobj is not None:
            return self.beam_focusing_hwobj.get_active_focus_mode()
        else:
            return "Collimated", None

    def set_focus_mode(self, mode):
        """Sets focusing mode

        :param mode: selected focusing mode
        :type mode: str
        """
        if self.beam_focusing_hwobj is not None:
            self.beam_focusing_hwobj.set_focus_mode(mode)

    def set_motor_focus_mode(self, motor, mode):
        """Sets focusing mode to a single motor

        :param motor: motor name
        :type motor: str
        :param mode: focusing mode
        :type mode: str
        """
        if self.beam_focusing_hwobj is not None:
            self.beam_focusing_hwobj.set_motor_focus_mode(motor, mode)

    def ppu_status_changed(self, is_error, text):
        """Reemits ppu status changed signal

        :param is_error: is error
        :type is_error: bool
        :param text: error message
        :type text: str
        """
        self.emit("ppuStatusChanged", (is_error, text))

    def ppu_restart_all(self):
        """Restart ppu processes"""
        if self.bl_hwobj.ppu_control_hwobj is not None:
            self.bl_hwobj.ppu_control_hwobj.restart_all()

    def test_ppu(self):
        """Test ppu"""
        result = {}
        if self.bl_hwobj.ppu_control_hwobj is not None:
            is_error, msg = self.bl_hwobj.ppu_control_hwobj.get_status()
            result["result_bit"] = not is_error
            if result["result_bit"]:
                result["result_short"] = "Test passed"
            else:
                result["result_short"] = "Test failed"

            msg = msg.replace("\n", "\n<br>")
            result["result_details"] = msg.split("\n")
        else:
            result["result_bit"] = False
            result["result_short"] = "Test failed (ppu hwobj not define)."

        self.ready_event.set()

        return result

    def test_focusing(self):
        """Tests focusing"""
        result = {}
        result["result_details"] = []

        active_mode, beam_size = self.get_focus_mode()
        if active_mode is None:
            result["result_bit"] = False
            result["result_short"] = "No focusing mode detected"
        else:
            result["result_bit"] = True
            result["result_short"] = "%s mode detected" % active_mode

        focus_modes = self.get_focus_mode_names()
        focus_motors_list = self.get_focus_motors()

        table_cells = []
        if focus_motors_list:
            for motor in focus_motors_list:
                table_row = []
                table_row.append(motor["motorName"])
                for focus_mode in focus_modes:
                    res = focus_mode in motor["focMode"]
                    table_row.append(
                        "<td bgcolor=%s>%.3f/%.3f</td>"
                        % (
                            TEST_COLORS_TABLE[res],
                            motor["focusingModes"][focus_mode],
                            motor["position"],
                        )
                    )
                table_cells.append(table_row)

        focus_modes = ["Motors"] + list(focus_modes)
        result["result_details"] = SimpleHTML.create_table(focus_modes, table_cells)

        self.ready_event.set()

        return result

    def measure_flux(self):
        """Measures intesity"""
        self.bl_hwobj.flux_hwobj.measure_flux()

    def stop_comm_process(self):
        """Stops pinging"""
        if self.current_test_procedure:
            self.current_test_procedure.kill()
            self.ready_event.set()

    def get_available_tests(self):
        """Returns a list with available tests"""
        return self.available_tests_dict

    def get_startup_test_list(self):
        """Returns a list with tests defined at startup"""
        test_list = []
        for test in self.startup_test_list:
            if TEST_DICT.get(test):
                test_list.append(TEST_DICT[test])
        return test_list

    def generate_report(self):
        """Generates html and pdf report"""
        html_filename = os.path.join(self.test_directory, self.test_filename) + ".html"
        pdf_filename = os.path.join(self.test_directory, self.test_filename) + ".pdf"
        archive_filename = os.path.join(
            self.test_directory,
            datetime.now().strftime("%Y_%m_%d_%H") + "_" + self.test_filename,
        )

        try:
            output_file = open(html_filename, "w")
            output_file.write(SimpleHTML.create_html_start("Beamline test summary"))
            output_file.write("<h1>Beamline %s Test results</h1>" % self.beamline_name)
            output_file.write("<h2>Executed tests:</h2>")
            table_cells = []
            for test in self.results_list:
                table_cells.append(
                    [
                        "bgcolor=%s" % TEST_COLORS_TABLE[test["result_bit"]],
                        "<a href=#%s>%s</a>" % (test["short_name"], test["full_name"]),
                        test["result_short"],
                        test["start_time"],
                        test["end_time"],
                    ]
                )

            table_rec = SimpleHTML.create_table(
                ["Name", "Result", "Start time", "End time"], table_cells
            )
            for row in table_rec:
                output_file.write(row)
            output_file.write("\n<hr>\n")

            for test_result in self.results_html_list:
                output_file.write(test_result + "\n")

            output_file.write(SimpleHTML.create_html_end())
            output_file.close()

            self.emit("htmlGenerated", html_filename)
            logging.getLogger("HWR").info(
                "BeamlineTest: Test result written in file %s" % html_filename
            )
        except Exception:
            logging.getLogger("HWR").error(
                "BeamlineTest: Unable to generate html report file %s" % html_filename
            )

        # try:
        #    pdfkit.from_url(html_filename, pdf_filename)
        #    logging.getLogger("GUI").info("PDF report %s generated" % pdf_filename)
        # except Exception:
        #    logging.getLogger("HWR").error(
        #        "BeamlineTest: Unable to generate pdf report file %s" % pdf_filename
        #    )

        self.emit("testFinished", html_filename)

    def get_result_html(self):
        """Returns html filename"""
        html_filename = os.path.join(self.test_directory, self.test_filename) + ".html"
        if os.path.exists(html_filename):
            return html_filename

    def test_method(self):
        print("Test")
